unit FRegistryBrowser;

{ FRegistryBrowser.pas
  FRegistryBrowser.dfm

Copyright (C) 2001  Richard T. Weeks

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Commercial development licenses for all files in the Registry Persistence
Package can be obtained from the author for US$10.  This includes
    RegistryPersist.pas
    FRegistryBrowser.pas
    FRegistryBrowser.dfm
    FPropertyPersist.pas
    FPropertyPersist.dfm
    FSimpleInput.pas
    FSimpleInput.dfm
    RegPersistDsgn.pas

e-mail addresses:
Richard T. Weeks - rtweeks@acm.org
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, ExtCtrls, StdCtrls, ActnList, Menus;

type
  TRegistryBrowserForm = class(TForm)
    Keys: TTreeView;
    Splitter1: TSplitter;
    Values: TListView;
    ButtonsPanel: TPanel;
    Button1: TButton;
    Button2: TButton;
    KeysPopupMenu: TPopupMenu;
    NewKey1: TMenuItem;
    Delete1: TMenuItem;
    ActionList1: TActionList;
    ANewKey: TAction;
    ADeleteKey: TAction;
    procedure KeysExpanding(Sender: TObject; Node: TTreeNode;
      var AllowExpansion: Boolean);
    procedure KeysCollapsed(Sender: TObject; Node: TTreeNode);
    procedure FormShow(Sender: TObject);
    procedure FormHide(Sender: TObject);
    procedure ANewKeyExecute(Sender: TObject);
    procedure ADeleteKeyExecute(Sender: TObject);
    procedure ADeleteKeyUpdate(Sender: TObject);
    procedure KeysChange(Sender: TObject; Node: TTreeNode);
    procedure FormPaint(Sender: TObject);
  private
    FRoot: HKEY;
    FUpdateThread: TThread;
    FFirstTime: Boolean;
    procedure SetRoot(const Value: HKEY);
    function GetRootName: String;
    procedure UpdateNode(Node: TTreeNode);
    procedure FreeNode(Node: TTreeNode);
    procedure UpdateNodes;
    function GetSelectedPath: String;
    procedure SetSelectedPath(const Value: String);
    procedure UpdateValues(Node: TTreeNode; SuspendUpdate: Boolean = True);
    { Private declarations }
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function ShowModal: Integer; override;
    property Root: HKEY read FRoot write SetRoot;
    property SelectedPath: String read GetSelectedPath write SetSelectedPath;
    { Public declarations }
  end;

var
  RegistryBrowserForm: TRegistryBrowserForm;

implementation

uses syncobjs, SimpleInput;

{$R *.DFM}

type
  TKeyData = class(TObject)
  private
    FEvent: DWORD;
    FKey: HKEY;
    procedure SetNotify;
  public
    constructor Create(AKey: HKEY);
    destructor Destroy; override;
    function Changed: Boolean;
    procedure Reset;
    property Key: HKEY read FKey;
    property Event: DWORD read FEvent;
  end;

  TUpdateThread = class(TThread)
  private
    FUpdateEvent: Cardinal;
    FForm: TRegistryBrowserForm;
    procedure SetUpdateNotify;
  protected
    procedure Execute; override;
  public
    constructor Create(AForm: TRegistryBrowserForm);
    destructor Destroy; override;
    procedure RootChanged;
  end;

{ TKeyData }

function TKeyData.Changed: Boolean;
begin
  Result := WaitForSingleObject(FEvent, 0) = WAIT_OBJECT_0;
end;

constructor TKeyData.Create(AKey: HKEY);
begin
  inherited Create;
  FKey := AKey;
  FEvent := CreateEvent(nil, True, False, nil);
  SetNotify;
end;

destructor TKeyData.Destroy;
begin
  CloseHandle(FEvent);
  RegCloseKey(FKey);
  inherited Destroy;
end;

procedure TKeyData.Reset;
begin
  ResetEvent(FEvent);
  SetNotify;
end;

procedure TKeyData.SetNotify;
begin
  RegNotifyChangeKeyValue(FKey, False, REG_NOTIFY_CHANGE_NAME or
    REG_NOTIFY_CHANGE_LAST_SET, FEvent, True);
end;

{ TUpdateThread }

constructor TUpdateThread.Create(AForm: TRegistryBrowserForm);
begin
  FForm := AForm;
  FUpdateEvent := CreateEvent(nil, True, False, nil);
  inherited Create(False);
end;

destructor TUpdateThread.Destroy;
begin
  Terminate;
  PulseEvent(FUpdateEvent);
  inherited Destroy;
end;

procedure TUpdateThread.Execute;
begin
  while not Terminated do begin
    SetUpdateNotify;
    WaitForSingleObject(FUpdateEvent, INFINITE);
    Synchronize(FForm.UpdateNodes);
  end;
end;

procedure TUpdateThread.RootChanged;
begin
  SetEvent(FUpdateEvent);
end;

procedure TUpdateThread.SetUpdateNotify;
begin
  ResetEvent(FUpdateEvent);
  RegNotifyChangeKeyValue(FForm.Root, True, REG_NOTIFY_CHANGE_NAME or
    REG_NOTIFY_CHANGE_LAST_SET, FUpdateEvent, True);
end;

{ TTRegistryBrowserForm }

constructor TRegistryBrowserForm.Create(AOwner: TComponent);
begin
  FFirstTime := True;
  inherited Create(AOwner);
  Root := HKEY_CURRENT_USER;
end;

function TRegistryBrowserForm.GetRootName: String;
begin
  case FRoot of
    HKEY_CURRENT_USER: Result := 'HKEY_CURRENT_USER';
    HKEY_CURRENT_CONFIG: Result := 'HKEY_CURRENT_CONFIG';
    HKEY_USERS: Result := 'HKEY_USERS';
    HKEY_LOCAL_MACHINE: Result := 'HKEY_LOCAL_MACHINE';
    HKEY_CLASSES_ROOT: Result := 'HKEY_CLASSES_ROOT';
    else Result := 'USER_ROOT';
  end;
end;

procedure TRegistryBrowserForm.SetRoot(const Value: HKEY);
var
  KeyData: TKeyData;
begin
  FRoot := Value;
  Keys.Items.Clear;
  Values.Items.Clear;
  KeyData := TKeyData.Create(FRoot);
  Keys.Items.AddObject(nil, GetRootName, KeyData);
  UpdateNode(Keys.Items.GetFirstNode);
  if FUpdateThread <> nil then
    TUpdateThread(FUpdateThread).RootChanged;
end;

function TRegistryBrowserForm.ShowModal: Integer;
begin
  ButtonsPanel.Visible := True;
  Result := inherited ShowModal;
  ButtonsPanel.Visible := False;
end;

function GetSubkeyName(Key: HKEY; Index: Integer; out Name: String): Boolean;
var
  EnumResult: Cardinal;
begin
  SetLength(Name, 256);
  EnumResult := RegEnumKey(Key, Index, PChar(Name), Length(Name));
  case EnumResult of
    ERROR_SUCCESS: SetLength(Name, StrLen(PChar(Name)));
    ERROR_NO_MORE_ITEMS: Name := '';
    else raise Exception.Create('Error enumerating subkeys');
  end;
  Result := EnumResult = ERROR_SUCCESS;
end;

procedure TRegistryBrowserForm.UpdateNode(Node: TTreeNode);
var
  OldNodeCount: Integer;

  function SubNodeByName(const Name: String): TTreeNode;
  var
    L, M, R, Comp: Integer;
  begin
    L := 0;
    R := OldNodeCount - 1;
    if Keys.SortType = stText then begin
      while L < R do begin
        M := (L + R) div 2;
        Result := Node[M];
        Comp := CompareText(Name, Result.Text);
        if Comp = 0 then
          Exit
        else if Comp < 0 then
          R := M - 1
        else
          L := M + 1;
      end;
      Result := Node[L];
      if Result.Text <> Name then
        Result := nil;
    end else begin
      for L := 0 to Node.Count - 1 do begin
        Result := Node[L];
        if Result.Text = Name then
          Exit;
      end;
      Result := nil;
    end;
  end;

var
  SubKeys: DWORD;
  I: Integer;
  NodeKey, SubKey: HKEY;
  SubkeyName: String;
  SubNode: TTreeNode;
begin
  NodeKey := TKeyData(Node.Data).Key;
  if Node.Expanded then begin
    for I := Node.Count - 1 downto 0 do begin
      if RegOpenKey(NodeKey, PChar(Node[I].Text), SubKey) = 0 then begin
        RegCloseKey(SubKey);
      end else begin
        FreeNode(Node[I]);
      end;
    end;

    OldNodeCount := Node.Count;
    I := 0;
    while True do begin
      if not GetSubkeyName(NodeKey, I, SubkeyName) then
        Break;
      Inc(I);

      SubNode := SubNodeByName(SubkeyName);
      if SubNode = nil then begin
        if RegOpenKey(NodeKey, PChar(SubkeyName), SubKey) = 0 then begin
          SubNode := Keys.Items.AddChildObject(Node, SubkeyName,
            TKeyData.Create(SubKey));
          UpdateNode(SubNode);
        end;
      end;
    end;

    Node.AlphaSort;
  end else begin
    RegQueryInfoKey(NodeKey, nil, nil, nil, @SubKeys, nil,
      nil, nil, nil, nil, nil, nil);
    Node.HasChildren := SubKeys > 0;
  end;

  if Node = Keys.Selected then
    UpdateValues(Node);
end;

procedure TRegistryBrowserForm.FreeNode(Node: TTreeNode);
begin
  while Node.Count > 0 do
    FreeNode(Node[0]);
  TKeyData(Node.Data).Free;
  Node.Free;
end;

procedure TRegistryBrowserForm.UpdateNodes;
var
  Node: TTreeNode;
begin
  Keys.Items.BeginUpdate;
  try
    Node := Keys.Items.GetFirstNode;
    while Node <> nil do begin
      with TKeyData(Node.Data) do begin
        if Changed then begin
          UpdateNode(Node);
          Reset;
        end;
      end;

      Node := Node.GetNext;
    end;
  finally
    Keys.Items.EndUpdate;
  end;
end;

destructor TRegistryBrowserForm.Destroy;
begin
  if Keys.Items.Count > 0 then
    FreeNode(Keys.Items.GetFirstNode);
  inherited Destroy;
end;

procedure TRegistryBrowserForm.KeysExpanding(Sender: TObject;
  Node: TTreeNode; var AllowExpansion: Boolean);
var
  I: Integer;
  NodeKey, SubKey: HKEY;
  SubkeyName: String;
  SubNode: TTreeNode;
  SubKeys: DWORD;
begin
  NodeKey := TKeyData(Node.Data).Key;

  RegQueryInfoKey(NodeKey, nil, nil, nil, @SubKeys, nil,
    nil, nil, nil, nil, nil, nil);
  AllowExpansion := SubKeys > 0;
  Node.HasChildren := AllowExpansion;
  if not AllowExpansion then
    Exit;

  I := 0;
  while True do begin
    if not GetSubkeyName(NodeKey, I, SubkeyName) then
      Break;
    Inc(I);

    if RegOpenKey(NodeKey, PChar(SubkeyName), SubKey) = 0 then begin
      SubNode := Node.Owner.AddChildObject(Node, SubkeyName,
        TKeyData.Create(SubKey));
      UpdateNode(SubNode);
    end;
  end;

  Node.AlphaSort;
end;

procedure TRegistryBrowserForm.KeysCollapsed(Sender: TObject;
  Node: TTreeNode);
begin
  while Node.Count > 0 do
    FreeNode(Node[0]);
  Node.HasChildren := True;
end;

procedure TRegistryBrowserForm.FormShow(Sender: TObject);
begin
  FUpdateThread := TUpdateThread.Create(Self);
end;

procedure TRegistryBrowserForm.FormHide(Sender: TObject);
begin
  FUpdateThread.Free;
end;

function TRegistryBrowserForm.GetSelectedPath: String;
var
  Node: TTreeNode;
begin
  Node := Keys.Selected;
  if Node = nil then
    Exit;

  while Node.Parent <> nil do begin
    Result := '\' + Node.Text + Result;
    Node := Node.Parent;
  end;

  if Result = '' then
    Result := '\';
end;

procedure TRegistryBrowserForm.SetSelectedPath(const Value: String);
begin

end;

procedure TRegistryBrowserForm.ANewKeyExecute(Sender: TObject);
var
  SubkeyName: String;
  SubKey: HKEY;
begin
  if Keys.Selected = nil then
    Exit;

  if GetValidatedInput('New Key', 'New subkey of '+SelectedPath+':', SubkeyName,
    nil) then
  begin
    if RegCreateKey(TKeyData(Keys.Selected.Data).Key, PChar(SubkeyName),
      SubKey) = 0 then
    begin
      UpdateNodes;
    end;
  end;
end;

procedure TRegistryBrowserForm.ADeleteKeyExecute(Sender: TObject);
var
  KeyName: String;
begin
  if (Keys.Selected = nil) or (Keys.Selected.Parent = nil) then
    Exit;

  KeyName := Keys.Selected.Text;
  if MessageDlg('Delete '+SelectedPath+'?', mtWarning, mbOkCancel, 0) =
    mrOk then
  begin
    if RegDeleteKey(TKeyData(Keys.Selected.Parent.Data).Key,
      PChar(KeyName)) = 0 then
    begin
      FreeNode(Keys.Selected);
    end;
  end;
end;

procedure TRegistryBrowserForm.ADeleteKeyUpdate(Sender: TObject);
begin
  ADeleteKey.Enabled := (Keys.Selected <> nil) and
    (Keys.Selected.Parent <> nil);
end;

procedure TRegistryBrowserForm.KeysChange(Sender: TObject;
  Node: TTreeNode);
begin
  UpdateValues(Node, not FFirstTime);
end;

procedure TRegistryBrowserForm.UpdateValues(Node: TTreeNode;
  SuspendUpdate: Boolean);
var
  NodeKey: HKEY;
  ValueName: String;
  ValueNameLen: Cardinal;
  I: Integer;
begin
  NodeKey := TKeyData(Node.Data).Key;
  if SuspendUpdate then
    Values.Items.BeginUpdate;
  try
    Values.Items.Clear;

    with Values.Items.Add do begin
      Caption := '[Default]';
    end;

    I := 0;
    while True do begin
      ValueName := '';
      SetLength(ValueName, 256);
      ValueNameLen := Length(ValueName);

      if RegEnumValue(NodeKey, I, PChar(ValueName), ValueNameLen,
        nil, nil, nil, nil) <> ERROR_SUCCESS then
        Break;

      SetLength(ValueName, ValueNameLen);
      with Values.Items.Add do begin
        Caption := ValueName;
      end;

      Inc(I);
    end;
  finally
    if SuspendUpdate then
      Values.Items.EndUpdate;
  end;
end;

procedure TRegistryBrowserForm.FormPaint(Sender: TObject);
begin
  FFirstTime := False;
end;

end.
